'''File for handling all operations relating to the Bundle resource'''

import uuid
from fhir.resources.bundle import Bundle


def generateBundle(bundle_entries: list, type: str = 'collection') -> dict:
    '''Generate Bundle Resource from resource entries'''
    if type == 'transaction':
        formatted_entries = [{
            'fullUrl': entry['resourceType'] + '/' + entry['id'], 'resource': entry, 'request': {'method': 'POST', 'url': entry['resourceType']}
        } for entry in bundle_entries]
    else:
        formatted_entries = [{
            'fullUrl': entry['resourceType'] + '/' + entry['id'], 'resource': entry
        } for entry in bundle_entries]

    bundle_data = {
        'id': str(uuid.uuid4()),
        'type': type,
        'entry': formatted_entries
    }

    bundle_resource = Bundle(**bundle_data)
    return dict(bundle_resource.dict())



'''File for handling all operations relating to the Observation resource'''

import uuid
import random
import datetime
from fhir.resources.observation import Observation, ObservationComponent
from fhir.resources.quantity import Quantity

from fhirgenerator.helpers.helpers import makeRandomDate


def generateObservation(resource_detail: dict, patient_id: str, start_date: str, days: int) -> dict:
    '''Generate Observation Resource from resource detail from configuration'''

    observation_id = str(uuid.uuid4())

    observation_code = random.choice(resource_detail['codes'])

    random_date = makeRandomDate(start_date, days)

    value_x_type, value_x_value = handleValueTypes(resource_detail, random_date)

    observation_data = {
        'id': observation_id,
        'status': 'final',
        'code': {
            'coding': [
                observation_code
            ]
        },
        'subject': {
            'reference': f'Patient/{patient_id}'
        },
        'effectiveDateTime': str(random_date),
        f'value{value_x_type}': value_x_value
    }

    if 'valueNone' in observation_data:
        del observation_data['valueNone']

    if 'profile' in resource_detail:
        observation_data['meta'] = {}
        observation_data['meta']['profile'] = resource_detail['profile']

    if 'components' in resource_detail:
        observation_data['component'] = []
        for component_details in resource_detail['components']:
            observation_data['component'].append(generateObservationComponent(component_details, observation_data['effectiveDateTime']))

    observation_resource = Observation(**observation_data).dict()
    return observation_resource


def generateObservationComponent(component_detail, effective_date_time: datetime = None):
    '''Generate a component for an Observation'''
    component_code = random.choice(component_detail['codes'])

    value_x_type, value_x_value = handleValueTypes(component_detail, effective_date_time)

    component_data = {
        'code': {
            'coding': [
                component_code
            ]
        },
        f'value{value_x_type}': value_x_value
    }

    component = ObservationComponent(**component_data).dict()
    return component


def handleValueTypes(detail, effective_date_time: datetime = None, decimal_value=None):
    '''Determine value[x] type for resource generation'''
    if 'enumSetList' in detail:
        enum_set_list = detail['enumSetList']

        if 'value' in enum_set_list[0]:
            value_x_type = 'Quantity'
            value_x_value = random.choice(enum_set_list)
        elif 'coding' in enum_set_list[0]:
            value_x_type = 'CodeableConcept'
            value_x_value = random.choice(enum_set_list)
        elif enum_set_list[0].isnumeric():
            value_x_type = 'Integer'
            value_x_value = random.choice(enum_set_list)
        elif len(enum_set_list[0].split(':')) > 1:
            value_x_type = 'Ratio'
            value_x_titer_choice = random.choice(enum_set_list)
            value_x_titer_choice_split = value_x_titer_choice.split(':')
            value_x_value = {
                'numerator': {'value': value_x_titer_choice_split[0]},
                'denominator': {'value': value_x_titer_choice_split[1]}
            }

        else:
            value_x_type = 'String'
            value_x_value = random.choice(enum_set_list)
    elif 'minValue' in detail and 'maxValue' in detail:
        # Quantity or Integer Value
        min_value = detail['minValue']
        max_value = detail['maxValue']
        if 'decimalValue' in detail:
            decimal_value = detail['decimalValue']

        if 'unit' in detail:
            value_x_type, value_x_value = createValueQuantity(min_value, max_value, detail['unit'], decimal_value)
        else:
            if decimal_value is not None:
                value_x_type, value_x_value = createValueQuantity(min_value, max_value, None, decimal_value)
            else:
                value_x_type, value_x_value = createValueInteger(min_value, max_value)
    elif 'dateRange' in detail:
        # DateTime Value
        if 'dateType' in detail:
            date_type = detail['dateType']
        else:
            date_type = 'datetime'
        value_x_type, value_x_value = createValueDateTime(effective_date_time, detail['dateRange'], date_type)
    else:
        print("Warning: There was no enumSetList or (minValue and maxValue) or dateRange in your configuration for this Observation. This Observation will not have a value[x].")
        value_x_type = 'None'
        value_x_value = ''

    return value_x_type, value_x_value


def createValueDateTime(effective_date_time: datetime, date_range: list, date_type: str = 'dateTime'):
    '''Generate a valueDateTime'''
    type_string = 'DateTime'
    offset = int(round(random.uniform(date_range[0], date_range[1])))
    value = effective_date_time + datetime.timedelta(days=offset)

    return type_string, value


def createValueInteger(min_value, max_value):
    '''Generate a valueInteger'''
    type_string = "Integer"
    value = int(round(random.uniform(min_value, max_value)))
    return type_string, value


def createValueQuantity(min_value, max_value, unit_coding=None, decimal=None):
    '''Generate a valueQuantity'''
    type_string = "Quantity"

    value = random.uniform(min_value, max_value)
    if decimal is not None:
        value = round(value, decimal)
    else:
        value = int(value)

    if unit_coding is not None:
        system, code, display = unit_coding.split('^')
        quantity_data = {
            'value': value,
            'unit': display,
            'system': system,
            'code': code
        }
    else:
        quantity_data = {
            'value': value
        }

    quantity = Quantity(**quantity_data).dict()
    return type_string, quantity



'''Testing Main function'''

import orjson
import datetime
from dateutil import parser
from fhirgenerator.main import generateResources
from fhirgenerator.helpers.helpers import default


def testMain():
    '''Testing function for main.py'''

    with open('fhirgenerator/tests/input/config.json', 'rb') as infile:
        config_dict = orjson.loads(infile.read())

    output_collection_bundle = generateResources(config_dict=config_dict, bundle_type='collection')
    output_trans_bundle = generateResources(config_dict=config_dict, bundle_type='transaction')

    with open('fhirgenerator/tests/output/test_main_transaction.json', 'wb') as outfile:
        outfile.write(orjson.dumps(output_trans_bundle, default=default, option=orjson.OPT_NAIVE_UTC))
    with open('fhirgenerator/tests/output/test_main_collection.json', 'wb') as outfile:
        outfile.write(orjson.dumps(output_collection_bundle, default=default, option=orjson.OPT_NAIVE_UTC))

    patient_list_collection = list(filter(lambda entry: entry['resource']['resourceType'] == 'Patient', output_collection_bundle['entry']))
    patient_list_transaction = list(filter(lambda entry: entry['resource']['resourceType'] == 'Patient', output_trans_bundle['entry']))

    observation_list_collection = list(filter(lambda entry: entry['resource']['resourceType'] == 'Observation', output_collection_bundle['entry']))
    observation_list_transaction = list(filter(lambda entry: entry['resource']['resourceType'] == 'Observation', output_trans_bundle['entry']))

    observation_dates_collection = [resource['resource']['effectiveDateTime'] for resource in observation_list_collection]
    observation_dates_transaction = [resource['resource']['effectiveDateTime'] for resource in observation_list_transaction]

    condition_list_collection = list(filter(lambda entry: entry['resource']['resourceType'] == 'Condition', output_collection_bundle['entry']))
    condition_list_transaction = list(filter(lambda entry: entry['resource']['resourceType'] == 'Condition', output_trans_bundle['entry']))

    condition_dates_collection = [resource['resource']['onsetDateTime'] for resource in condition_list_collection]
    condition_dates_transaction = [resource['resource']['onsetDateTime'] for resource in condition_list_transaction]

    obs_mins = [detail['minOccurrencesPerCycle'] if detail['fhirResource'] == 'Observation' else 0 for detail in config_dict['resourceDetails']]
    obs_maxs = [detail['maxOccurrencesPerCycle'] if detail['fhirResource'] == 'Observation' else 0 for detail in config_dict['resourceDetails']]
    min_num_of_observations = config_dict['numberPatients'] * sum(obs_mins)
    max_num_of_observations = config_dict['numberPatients'] * sum(obs_maxs)

    con_mins = [detail['minOccurrencesPerCycle'] if detail['fhirResource'] == 'Condition' else 0 for detail in config_dict['resourceDetails']]
    con_maxs = [detail['maxOccurrencesPerCycle'] if detail['fhirResource'] == 'Condition' else 0 for detail in config_dict['resourceDetails']]
    min_num_of_conditions = config_dict['numberPatients'] * sum(con_mins)
    max_num_of_conditions = config_dict['numberPatients'] * sum(con_maxs)

    start_date = parser.parse(config_dict['startDate'])
    end_date = start_date + datetime.timedelta(days=config_dict['days'])

    obs_dates_in_range_collection = [date_time >= start_date and date_time <= end_date for date_time in observation_dates_collection]
    obs_dates_in_range_transaction = [date_time >= start_date and date_time <= end_date for date_time in observation_dates_transaction]

    con_dates_in_range_collection = [date_time >= start_date and date_time <= end_date for date_time in condition_dates_collection]
    con_dates_in_range_transaction = [date_time >= start_date and date_time <= end_date for date_time in condition_dates_transaction]

    assert len(patient_list_collection) == config_dict['numberPatients']
    assert len(observation_list_collection) >= min_num_of_observations and len(observation_list_collection) <= max_num_of_observations
    assert all(obs_dates_in_range_collection)
    assert len(condition_list_collection) >= min_num_of_conditions and len(condition_list_collection) <= max_num_of_conditions
    assert all(con_dates_in_range_collection)

    assert len(patient_list_transaction) == config_dict['numberPatients']
    assert len(observation_list_transaction) >= min_num_of_observations and len(observation_list_transaction) <= max_num_of_observations
    assert all(obs_dates_in_range_transaction)
    assert len(condition_list_transaction) >= min_num_of_conditions and len(condition_list_transaction) <= max_num_of_conditions
    assert all(con_dates_in_range_transaction)


def testMainUSCore():
    '''Testing function for generateResources with US Core Config'''

    with open('fhirgenerator/tests/input/config_usCore.json', 'rb') as infile:
        config_dict = orjson.loads(infile.read())

    output_collection_bundle = generateResources(config_dict=config_dict, bundle_type='collection')
    output_trans_bundle = generateResources(config_dict=config_dict, bundle_type='transaction')

    with open('fhirgenerator/tests/output/test_main_uscore_transaction.json', 'wb') as outfile:
        outfile.write(orjson.dumps(output_trans_bundle, default=default, option=orjson.OPT_NAIVE_UTC))
    with open('fhirgenerator/tests/output/test_main_uscore_collection.json', 'wb') as outfile:
        outfile.write(orjson.dumps(output_collection_bundle, default=default, option=orjson.OPT_NAIVE_UTC))

    patient_list_collection = list(filter(lambda entry: entry['resource']['resourceType'] == 'Patient', output_collection_bundle['entry']))
    patient_list_transaction = list(filter(lambda entry: entry['resource']['resourceType'] == 'Patient', output_trans_bundle['entry']))

    assert len(patient_list_collection) == config_dict['numberPatients']
    assert len(patient_list_transaction) == config_dict['numberPatients']
    # TODO: Add more assertions for US Core Bundles



'''Testing MedicationStatement generator'''

import orjson
import pytest
from dateutil import parser
import datetime
from fhirgenerator.resources.r4.medicationStatement import generateMedicationStatement
from fhirgenerator.helpers.helpers import default


def testMedicationStatementGenerator():
    '''Test function for MedicationStatement Generator'''

    with open('fhirgenerator/tests/input/config.json', 'rb') as infile:
        config_dict = orjson.loads(infile.read())

    patient_id = '26774-827647-736278-3737646'

    ms_resource_details = []
    for detail in config_dict['resourceDetails']:
        if detail['fhirResource'] == 'MedicationStatement':
            ms_resource_details.append(detail)

    for i, detail in enumerate(ms_resource_details):
        created_resource = generateMedicationStatement(detail, patient_id, config_dict['startDate'], config_dict['days'])

        assert created_resource['resourceType'] == 'MedicationStatement'
        assert created_resource['status'] == 'completed'
        assert created_resource['medicationCodeableConcept']['coding'][0] in detail['codes']
        assert (created_resource['effectiveDateTime'] >= parser.parse(config_dict['startDate'])) and (created_resource['effectiveDateTime'] <= parser.parse(config_dict['startDate']) + datetime.timedelta(days=config_dict['days']))
        assert created_resource['subject']['reference'] == f'Patient/{patient_id}'

        dose_quantity = created_resource['dosage'][0]['doseAndRate'][0]['doseQuantity']
        if 'enumSetList' in detail:
            enum_set_list = detail['enumSetList']
            assert str(dose_quantity['value']) + ' ' + dose_quantity['system'] + '^' + dose_quantity['code'] + '^' + dose_quantity['unit'] in enum_set_list
        else:
            assert dose_quantity['value'] <= detail['maxValue'] and dose_quantity['value'] >= detail['minValue']
            test_system, test_code, test_unit = detail['unit'].split('^')
            assert dose_quantity['system'] == test_system
            assert dose_quantity['code'] == test_code
            assert dose_quantity['unit'] == test_unit

        with open(f'fhirgenerator/tests/output/test_medicationStatement_{i}.json', 'wb') as outfile:
            outfile.write(orjson.dumps(created_resource, default=default, option=orjson.OPT_NAIVE_UTC))


def testMedicationStatementWithError():
    '''Test function for MedicationStatement Generator that should raise a NameError'''

    patient_id = '26774-827647-736278-3737646'

    detail = {
        "fhirResource": "MedicationStatement",
        "codes": [
            {"system": "http://www.nlm.nih.gov/research/umls/rxnorm", "code": "18631"},
            {"system": "http://www.nlm.nih.gov/research/umls/rxnorm", "code": "3640"},
            {"system": "http://www.nlm.nih.gov/research/umls/rxnorm", "code": "8698"}
        ],
        "minValue": 3,
        "maxValue": 7,
        "minOccurrencesPerCycle": 1,
        "maxOccurrencesPerCycle": 8,
        "cycleLengthInDays": 365
    }
    with pytest.raises(NameError):
        generateMedicationStatement(detail, patient_id, '01-01-2022', 365)



'''File for handling all operations relating to the US Core Location resource'''

from fhirgenerator.resources.r4.location import generateLocation


def generateUSCoreLocation(config: dict) -> dict:
    '''Generate US Core Location Resource'''

    location_resource = generateLocation()
    location_resource['meta'] = {'profile': ['http://hl7.org/fhir/us/core/StructureDefinition/us-core-location']}

    return location_resource



'''File for handling US Core Resource Generation'''

from fhirgenerator.resources.uscore_r4.usCoreBMIObservation import generateUSCoreBMIObservation
from fhirgenerator.resources.uscore_r4.usCoreBodyHeightObservation import generateUSCoreBodyHeightObservation
from fhirgenerator.resources.uscore_r4.usCoreBodyTemperatureObservation import generateUSCoreBodyTemperatureObservation
from fhirgenerator.resources.uscore_r4.usCoreBodyWeightObservation import generateUSCoreBodyWeightObservation
from fhirgenerator.resources.uscore_r4.usCoreCondition import generateUSCoreCondition
from fhirgenerator.resources.uscore_r4.usCoreHeadCircumferenceObservation import generateUSCoreHeadCircumferenceObservation
from fhirgenerator.resources.uscore_r4.usCoreHeartRateObservation import generateUSCoreHeartRateObservation
from fhirgenerator.resources.uscore_r4.usCoreLabResultObservation import generateUSCoreLabResultObservation
from fhirgenerator.resources.uscore_r4.usCoreLocation import generateUSCoreLocation


def handleUSCore(resource_detail: dict, patient_id: str, start_date: str, days: str) -> dict:
    '''Function to handle US Core resource generation'''

    resource_type = resource_detail['fhirResource']
    match resource_type:
        case 'USCoreCondition':
            return_resource = generateUSCoreCondition(detail=resource_detail, patient_id=patient_id, start_date=start_date, days=days)
        case 'USCoreLabResultObservation':
            return_resource = generateUSCoreLabResultObservation(detail=resource_detail, patient_id=patient_id, start_date=start_date, days=days)
        case 'USCoreBMIObservation':
            return_resource = generateUSCoreBMIObservation(detail=resource_detail, patient_id=patient_id, start_date=start_date, days=days)
        case 'USCoreHeadCircumferenceObservation':
            return_resource = generateUSCoreHeadCircumferenceObservation(detail=resource_detail, patient_id=patient_id, start_date=start_date, days=days)
        case 'USCoreBodyHeightObservation':
            return_resource = generateUSCoreBodyHeightObservation(detail=resource_detail, patient_id=patient_id, start_date=start_date, days=days)
        case 'USCoreBodyWeightObservation':
            return_resource = generateUSCoreBodyWeightObservation(detail=resource_detail, patient_id=patient_id, start_date=start_date, days=days)
        case 'USCoreBodyTemperatureObservation':
            return_resource = generateUSCoreBodyTemperatureObservation(detail=resource_detail, patient_id=patient_id, start_date=start_date, days=days)
        case 'USCoreHeartRateObservation':
            return_resource = generateUSCoreHeartRateObservation(detail=resource_detail, patient_id=patient_id, start_date=start_date, days=days)
        case 'USCoreLocation':
            return_resource = generateUSCoreLocation()

    return return_resource



'''Testing Location generator'''

import orjson
from fhirgenerator.resources.r4.location import generateLocation
from fhirgenerator.helpers.helpers import default
from fhir.resources.address import Address
from fhir.resources.contactpoint import ContactPoint


def testLocationGenerator():
    '''Test function for Location Generator'''

    with open('fhirgenerator/tests/input/config.json', 'rb') as infile:
        config_dict = orjson.loads(infile.read())

    location_resource_details = []
    for detail in config_dict['resourceDetails']:
        if detail['fhirResource'] == 'Location':
            location_resource_details.append(detail)

    for i, detail in enumerate(location_resource_details):

        created_resource = generateLocation()

        assert created_resource['resourceType'] == 'Location'
        assert created_resource['status'] == 'active'

        assert isinstance(Address(**created_resource['address']), Address)
        assert isinstance(ContactPoint(**created_resource['telecom'][0]), ContactPoint)

        with open(f'fhirgenerator/tests/output/test_location_{i}.json', 'wb') as outfile:
            outfile.write(orjson.dumps(created_resource, default=default, option=orjson.OPT_NAIVE_UTC))



'''File for handling all operations relating to the Location resource'''

import uuid
from faker import Faker
from fhir.resources.location import Location
from fhirgenerator.types.address import generateAddress

from fhirgenerator.types.contactPoint import generateContactPoint


def generateLocation() -> dict:
    '''Generate a Location resource'''

    fake = Faker()

    location_id = str(uuid.uuid4())

    location_dict = {
        'id': location_id,
        'name': fake.company(),
        'status': 'active',
        'telecom': [generateContactPoint()],
        'address': generateAddress(use='work')
    }

    location = Location(**location_dict)

    return dict(location.dict())



'''Testing US Core Patient generator'''

import orjson
from fhirgenerator.resources.uscore_r4.usCorePatient import generateUSCorePatient
from fhirgenerator.helpers.helpers import calculateAge, default


def testUSCorePatientGenerator():
    '''Test function for US Core Patient Generator'''
    patient_config = {
        "age": 66,
        "gender": 'F',
        "startDate": '01-01-2022'
    }
    created_resource = generateUSCorePatient(patient_config)

    assert created_resource['resourceType'] == 'Patient'
    assert created_resource['meta']['profile'][0] == 'http://hl7.org/fhir/us/core/StructureDefinition/us-core-patient'
    assert created_resource['extension'][0]['url'] in ['http://hl7.org/fhir/us/core/StructureDefinition/us-core-race', 'http://hl7.org/fhir/us/core/StructureDefinition/us-core-ethnicity']
    assert created_resource['extension'][1]['url'] in ['http://hl7.org/fhir/us/core/StructureDefinition/us-core-race', 'http://hl7.org/fhir/us/core/StructureDefinition/us-core-ethnicity']
    assert created_resource['gender'] == 'female'
    assert calculateAge(created_resource['birthDate'], start=patient_config['startDate']) == patient_config['age']
    assert isinstance(created_resource['name'], list)
    assert isinstance(created_resource['address'], list)
    assert created_resource['identifier'][0]['system'] == 'urn:fhirgen:mrn'
    assert created_resource['identifier'][0]['type']['coding'][0] == {'system': 'http://terminology.hl7.org/CodeSystem/v2-0203', 'code': 'MR'}
    assert len(created_resource['identifier'][0]['value']) == 11

    with open('fhirgenerator/tests/output/test_usCorePatient.json', 'wb') as outfile:
        outfile.write(orjson.dumps(created_resource, default=default, option=orjson.OPT_NAIVE_UTC))



'''Command line functionality for package to generate new Python files for extending this package'''

import argparse
import os


def main():
    '''
    Command line function to generate new Python files for extending this package
    example usage: fhirgenerator new profile observation ObservationCauseOfDeathCondition
    '''
    parser = argparse.ArgumentParser(prog='fhirgenerator_make_new_profile')
    parser.add_argument('action', help='Only supports new at this time')
    parser.add_argument('type', help='Only supports profile at this time')
    parser.add_argument('resource_type', help='name of resource template you want to generate')
    parser.add_argument('profile_name', help='name of profile you want to generate')

    args = parser.parse_args()

    assert args.action == 'new'
    assert args.type == 'profile'

    resource_type = args.resource_type
    profile_name = args.profile_name
    output_file_text = f'''from fhirgenerator.resources.r4.{resource_type.lower()} import generate{resource_type.capitalize()}
from fhir.resources.{resource_type.lower()} import {resource_type.capitalize()}


def generate{profile_name[:1].upper() + profile_name[1:]}(resource_detail, patient_id, start_date, days):
    # Pre-built template using the fhirgenerator command line tool

    {resource_type.lower()} = {{}}

    # START CODE

    # END CODE
    {resource_type.lower()} = generate{resource_type.capitalize()}(resource_detail, patient_id, start_date, days)

    {resource_type.lower()} = {resource_type.capitalize()}(**{resource_type.lower()}).dict()
    return {resource_type.lower()}
'''

    if os.path.isfile(f'generate{profile_name}.py'):
        overwrite = input('File already exists. Overwrite (Y/n)? ')
        if overwrite.lower() == 'y':
            with open(f'generate{profile_name}.py', 'w') as outfile:
                outfile.write(output_file_text)
            print(f'Saved new Profile file at generate{profile_name}.py')
        else:
            print('File not saved')
    else:
        with open(f'generate{profile_name}.py', 'w') as outfile:
            outfile.write(output_file_text)
        print(f'Saved new Profile file at generate{profile_name}.py')



'''File for handling all operations relating to the Encounter resource'''

import uuid
import random
from fhir.resources.encounter import Encounter

from fhirgenerator.helpers.helpers import makeRandomDate

class_options = [
    {
        'system': 'http://terminology.hl7.org/CodeSystem/v3-ActCode',
        'code': 'AMB',
        'display': 'ambulatory'
    },
    {
        'system': 'http://terminology.hl7.org/CodeSystem/v3-ActCode',
        'code': 'EMER',
        'display': 'emergency'
    },
    {
        'system': 'http://terminology.hl7.org/CodeSystem/v3-ActCode',
        'code': 'FLD',
        'display': 'field'
    },
    {
        'system': 'http://terminology.hl7.org/CodeSystem/v3-ActCode',
        'code': 'HH',
        'display': 'home health'
    },
    {
        'system': 'http://terminology.hl7.org/CodeSystem/v3-ActCode',
        'code': 'IMP',
        'display': 'inpatient encounter'
    },
    {
        'system': 'http://terminology.hl7.org/CodeSystem/v3-ActCode',
        'code': 'ACUTE',
        'display': 'inpatient acute'
    },
    {
        'system': 'http://terminology.hl7.org/CodeSystem/v3-ActCode',
        'code': 'NONAC',
        'display': 'inpatient non-acute'
    },
    {
        'system': 'http://terminology.hl7.org/CodeSystem/v3-ActCode',
        'code': 'OBSENC',
        'display': 'observation encounter'
    },
    {
        'system': 'http://terminology.hl7.org/CodeSystem/v3-ActCode',
        'code': 'PRENC',
        'display': 'pre-admission'
    },
    {
        'system': 'http://terminology.hl7.org/CodeSystem/v3-ActCode',
        'code': 'SS',
        'display': 'short-stay'
    },
    {
        'system': 'http://terminology.hl7.org/CodeSystem/v3-ActCode',
        'code': 'VR',
        'display': 'virtual'
    }
]


def generateEncounter(resource_detail: dict, patient_id: str, start_date: str, days: int) -> dict:
    '''Generate Encounter resource from resource detail from configuration'''

    encounter_id = str(uuid.uuid4())

    encounter_code = random.choice(resource_detail['codes'])

    if 'class' in resource_detail:
        encounter_class = resource_detail['class']
    else:
        encounter_class = random.choice(class_options)

    random_date = makeRandomDate(start_date, days)

    encounter_data = {
        'id': encounter_id,
        'status': 'finished',
        'class': encounter_class,
        'type': [{'coding': [encounter_code]}],
        'subject': {
            'reference': f'Patient/{patient_id}'
        },
        'period': {
            'start': random_date,
            'end': makeRandomDate(str(random_date), 10)
        }
    }

    encounter_resource = Encounter(**encounter_data).dict()
    return encounter_resource



'''File for handling all operations relating to the Dosage datatype'''

import random


def generateDosage(enum_set_list: list = [], min_value: int = 0, max_value: int = 0, decimal_value: int = 0, unit_str: str = '') -> dict:
    '''Generate a Dosage for a MedicationStatement'''
    if enum_set_list:
        dose_quantity_full = random.choice(enum_set_list)
        value, unit_string = dose_quantity_full.split(' ')
        dose_quantity_system, dose_quantity_code, dose_quantity_unit = unit_string.split('^')
        dose_quantity = {
            'value': value,
            'system': dose_quantity_system,
            'code': dose_quantity_code,
            'unit': dose_quantity_unit
        }
    else:
        if not unit_str:
            raise NameError('There was no enumSetList in your configuration and there is no listed unit. Without an enumSetList, you MUST have the unit key in your configuration')
        if decimal_value:
            system, code, unit = unit_str.split('^')
            dose_quantity = {
                'value': round(random.uniform(min_value, max_value), decimal_value),
                'system': system,
                'code': code,
                'unit': unit
            }
        else:
            system, code, unit = unit_str.split('^')
            dose_quantity = {
                'value': round(random.uniform(min_value, max_value)),
                'system': system,
                'code': code,
                'unit': unit
            }

    dosage_dict = {
        'doseAndRate': [{
            'doseQuantity': dose_quantity
        }]
    }
    return dosage_dict



'''Testing Observation generator'''

import orjson
from dateutil import parser
import datetime
from fhirgenerator.resources.r4.observation import generateObservation
from fhirgenerator.helpers.helpers import default


def testObservationGenerator():
    '''Test function for Observation Generator'''

    with open('fhirgenerator/tests/input/config.json', 'rb') as infile:
        config_dict = orjson.loads(infile.read())

    patient_id = '26774-827647-736278-3737646'

    observation_resource_details = []
    for detail in config_dict['resourceDetails']:
        if detail['fhirResource'] == 'Observation':
            observation_resource_details.append(detail)

    for i, detail in enumerate(observation_resource_details):

        if 'enumSetList' in detail:
            enum_set_list = detail['enumSetList']
            if 'value' in enum_set_list[0]:
                value_x_type = 'Quantity'
            elif 'coding' in enum_set_list[0]:
                value_x_type = 'CodeableConcept'
            elif len(enum_set_list[0].split(':')) > 1:
                value_x_type = 'Ratio'
            else:
                value_x_type = 'String'
        elif 'decimalValue' in detail:
            value_x_type = 'Quantity'
        elif 'minValue' in detail:
            value_x_type = 'Integer'
        elif 'dateRange' in detail:
            value_x_type = 'DateTime'
        else:
            value_x_type = 'None'

        created_resource = generateObservation(detail, patient_id, config_dict['startDate'], config_dict['days'])

        assert created_resource['resourceType'] == 'Observation'
        assert created_resource['status'] == 'final'
        assert created_resource['code']['coding'][0] in detail['codes']
        assert (created_resource['effectiveDateTime'] >= parser.parse(config_dict['startDate'])) and (created_resource['effectiveDateTime'] <= parser.parse(config_dict['startDate']) + datetime.timedelta(days=config_dict['days']))
        assert created_resource['subject']['reference'] == f'Patient/{patient_id}'

        match value_x_type:
            case 'Ratio':
                assert str(created_resource['valueRatio']['numerator']['value']) + ':' + str(created_resource['valueRatio']['denominator']['value']) in detail['enumSetList']
            case 'CodeableConcept':
                assert created_resource['valueCodeableConcept'] in detail['enumSetList']
            case 'Quantity':
                if 'enumSetList' in detail:
                    assert created_resource['valueQuantity'] in detail['enumSetList']
                else:
                    assert created_resource['valueQuantity']['value'] <= detail['maxValue'] and created_resource['valueQuantity']['value'] >= detail['minValue']
                    if 'unit' in detail:
                        test_system, test_code, test_unit = detail['unit'].split('^')
                        assert created_resource['valueQuantity']['system'] == test_system
                        assert created_resource['valueQuantity']['code'] == test_code
                        assert created_resource['valueQuantity']['unit'] == test_unit
            case 'String':
                assert created_resource['valueString'] in detail['enumSetList']
            case 'Integer':
                if 'enumSetList' in detail:
                    assert created_resource['valueInteger'] in detail['enumSetList']
                else:
                    assert created_resource['valueInteger'] <= detail['maxValue'] and created_resource['valueInteger'] >= detail['minValue']
            case 'DateTime':
                assert created_resource['valueDateTime'] >= created_resource['effectiveDateTime'] + datetime.timedelta(days=detail["dateRange"][0])
                assert created_resource['valueDateTime'] <= created_resource['effectiveDateTime'] + datetime.timedelta(days=detail["dateRange"][1])
            case 'None':
                assert 'valueRatio' not in created_resource
                assert 'valueCodeableConcept' not in created_resource
                assert 'valueQuantity' not in created_resource
                assert 'valueString' not in created_resource
                assert 'valueInteger' not in created_resource

        with open(f'fhirgenerator/tests/output/test_observation_{i}.json', 'wb') as outfile:
            outfile.write(orjson.dumps(created_resource, default=default, option=orjson.OPT_NAIVE_UTC))


def testObservationComponentGenerator():
    '''Test Observation Component Generator'''

    with open('fhirgenerator/tests/input/config.json', 'rb') as infile:
        config_dict = orjson.loads(infile.read())

    resource_details = {
        "fhirResource": "Observation",
        "profile": ["http://hl7.org/fhir/us/core/StructureDefinition/us-core-blood-pressure"],
        "codes": [
            {"system": "http://loinc.org", "code": "85354-9", "display": "Blood pressure panel with all children optional"}
        ],
        "minOccurrencesPerCycle": 1,
        "maxOccurrencesPerCycle": 1,
        "cycleLengthInDays": 365,
        "components": [
            {
                "codes": [{"system": "http://loinc.org", "code": "8480-6", "display": "Systolic blood pressure"}],
                "minValue": 105,
                "maxValue": 135,
                "unit": "http://unitsofmeasure.org^mm[Hg]^mmHg"
            },
            {
                "codes": [{"system": "http://loinc.org", "code": "8462-4", "display": "Diastolic blood pressure"}],
                "minValue": 65,
                "maxValue": 95,
                "unit": "http://unitsofmeasure.org^mm[Hg]^mmHg"
            }
        ]
    }

    patient_id = '26774-827647-736278-3737646'

    created_resource = generateObservation(resource_details, patient_id, config_dict['startDate'], config_dict['days'])

    with open('fhirgenerator/tests/output/test_observation_component.json', 'wb') as outfile:
        outfile.write(orjson.dumps(created_resource, default=default, option=orjson.OPT_NAIVE_UTC))


# TODO: write more observation tests for all possible value[x] that the package supports



'''File for handling all operations relating to the US Core Laboratory Result Observation resource'''

from fhirgenerator.resources.r4.observation import generateObservation


def generateUSCoreLabResultObservation(detail: dict, patient_id: str, start_date: str, days: str) -> dict:
    '''Generate a US Core Lab Result Observation'''

    observation_resource = generateObservation(detail, patient_id, start_date, days)
    observation_resource['meta'] = {'profile': ['http://hl7.org/fhir/us/core/StructureDefinition/us-core-observation-lab']}
    observation_resource['category'] = [{"coding": [{
        "system": "http://terminology.hl7.org/CodeSystem/observation-category",
        "code": "laboratory",
        "display": "Laboratory"
    }]}]
    return observation_resource



'''Main entrypoint for package'''

from random import randint

from fhirgenerator.resources.r4.encounter import generateEncounter
from fhirgenerator.resources.r4.patient import generatePatient
from fhirgenerator.resources.r4.observation import generateObservation
from fhirgenerator.resources.r4.condition import generateCondition
from fhirgenerator.resources.r4.bundle import generateBundle
from fhirgenerator.resources.r4.medicationStatement import generateMedicationStatement
from fhirgenerator.resources.r4.procedure import generateProcedure
from fhirgenerator.resources.r4.location import generateLocation

from fhirgenerator.resources.uscore_r4.handleUSCore import handleUSCore
from fhirgenerator.resources.uscore_r4.usCorePatient import generateUSCorePatient


def generateResources(config_dict: dict, bundle_type: str = 'collection') -> dict:
    '''Main function for generating resources'''

    final_bundle_entries = []

    # Getting total numbers of each gender
    gender_totals = [round(value * config_dict['numberPatients'] * .01) for value in config_dict['genderMFOU']]
    if sum(gender_totals) > config_dict['numberPatients']:
        gender_totals[0] -= (sum(gender_totals) - config_dict['numberPatients'])
    elif sum(gender_totals) < config_dict['numberPatients']:
        gender_totals[2] += (config_dict['numberPatients'] - sum(gender_totals))

    gender_totals_list = []
    for i in range(0, gender_totals[0]):
        gender_totals_list.append('M')
    for i in range(0, gender_totals[1]):
        gender_totals_list.append('F')
    for i in range(0, gender_totals[2]):
        gender_totals_list.append('O')
    for i in range(0, gender_totals[3]):
        gender_totals_list.append('U')

    for j in range(0, config_dict['numberPatients']):

        patient_age = randint(config_dict['ageMin'], config_dict['ageMax'])
        patient_config = {
            "age": patient_age,
            "gender": gender_totals_list[j],
            "startDate": config_dict['startDate']
        }
        if 'usCorePatient' in config_dict and config_dict['usCorePatient'] is True:
            patient_resource = generateUSCorePatient(patient_config)
        else:
            patient_resource = generatePatient(patient_config)
        final_bundle_entries.append(patient_resource)

        for resource_detail in config_dict['resourceDetails']:
            bundle_entry_list = []
            num_of_cycles = round(resource_detail['cycleLengthInDays'] / config_dict['days'])
            num_of_resources = randint(resource_detail['minOccurrencesPerCycle'], resource_detail['maxOccurrencesPerCycle']) * num_of_cycles
            resource_type = resource_detail['fhirResource']

            if (resource_type.lower()[0:6]) == 'uscore':
                for k in range(0, num_of_resources):
                    bundle_entry_list.append(handleUSCore(resource_detail, patient_resource['id'], config_dict['startDate'], config_dict['days']))

            match resource_type:
                case 'Observation':
                    for k in range(0, num_of_resources):
                        bundle_entry_list.append(generateObservation(resource_detail, patient_resource['id'], config_dict['startDate'], config_dict['days']))
                case 'Condition':
                    for k in range(0, num_of_resources):
                        bundle_entry_list.append(generateCondition(resource_detail, patient_resource['id'], config_dict['startDate'], config_dict['days']))
                case 'Procedure':
                    for k in range(0, num_of_resources):
                        bundle_entry_list.append(generateProcedure(resource_detail, patient_resource['id'], config_dict['startDate'], config_dict['days']))
                case 'MedicationStatement':
                    for k in range(0, num_of_resources):
                        bundle_entry_list.append(generateMedicationStatement(resource_detail, patient_resource['id'], config_dict['startDate'], config_dict['days']))
                case 'Encounter':
                    for k in range(0, num_of_resources):
                        bundle_entry_list.append(generateEncounter(resource_detail, patient_resource['id'], config_dict['startDate'], config_dict['days']))
                case 'Location':
                    for k in range(0, num_of_resources):
                        bundle_entry_list.append(generateLocation())

            final_bundle_entries.extend(bundle_entry_list)

        print(f'Patient number {j+1} generated')

    if 'bundleType' in config_dict:
        bundle_type = config_dict['bundleType']

    final_bundle = generateBundle(final_bundle_entries, type=bundle_type)

    return final_bundle



'''Testing US Core Heart Rate Observation generator'''

import orjson
from dateutil import parser
import datetime
from fhirgenerator.resources.uscore_r4.usCoreHeartRateObservation import generateUSCoreHeartRateObservation
from fhirgenerator.helpers.helpers import default


def testUSCoreHeartRateObservationGenerator():
    '''Test function for US Core Heart Rate Observation Generator'''

    with open('fhirgenerator/tests/input/config_usCore.json', 'rb') as infile:
        config_dict = orjson.loads(infile.read())

    patient_id = '26774-827647-736278-3737646'

    observation_resource_details = []
    for detail in config_dict['resourceDetails']:
        if detail['fhirResource'] == 'USCoreHeartRateObservation':
            observation_resource_details.append(detail)

    for i, detail in enumerate(observation_resource_details):
        created_resource = generateUSCoreHeartRateObservation(detail, patient_id, config_dict['startDate'], config_dict['days'])

        assert created_resource['resourceType'] == 'Observation'
        assert created_resource['meta']['profile'] == ['http://hl7.org/fhir/us/core/StructureDefinition/us-core-heart-rate']
        assert created_resource['status'] == 'final'
        assert created_resource['category'] == [{"coding": [{
            "system": "http://terminology.hl7.org/CodeSystem/observation-category",
            "code": "vital-signs",
            "display": "Vital Signs"
        }]}]
        assert created_resource['code']['coding'][0] == {'system': 'http://loinc.org', 'code': '8867-4'}
        assert (created_resource['effectiveDateTime'] >= parser.parse(config_dict['startDate'])) and (created_resource['effectiveDateTime'] <= parser.parse(config_dict['startDate']) + datetime.timedelta(days=config_dict['days']))
        assert created_resource['subject']['reference'] == f'Patient/{patient_id}'
        assert 'value' in created_resource['valueQuantity']
        assert 'unit' in created_resource['valueQuantity']
        assert 'system' in created_resource['valueQuantity']
        assert 'code' in created_resource['valueQuantity']

        with open(f'fhirgenerator/tests/output/test_usCoreHeartRateObservation_{i}.json', 'wb') as outfile:
            outfile.write(orjson.dumps(created_resource, default=default, option=orjson.OPT_NAIVE_UTC))



'''File for handling all operations relating to the US Core Patient resource'''

import random
from fhirgenerator.resources.r4.patient import generatePatient
from fhir.resources.extension import Extension

race_choices = [
    {
        'system': 'urn:oid:2.16.840.1.113883.6.238',
        'code': '1002-5',
        'display': 'American Indian or Alaska Native'
    },
    {
        'system': 'urn:oid:2.16.840.1.113883.6.238',
        'code': '2028-9',
        'display': 'Asian'
    },
    {
        'system': 'urn:oid:2.16.840.1.113883.6.238',
        'code': '2054-5',
        'display': 'Black or African American'
    },
    {
        'system': 'urn:oid:2.16.840.1.113883.6.238',
        'code': '2076-8',
        'display': 'Native Hawaiian or Other Pacific Islander'
    },
    {
        'system': 'urn:oid:2.16.840.1.113883.6.238',
        'code': '2106-3',
        'display': 'White'
    },
    {
        'system': 'http://terminology.hl7.org/CodeSystem/v3-NullFlavor',
        'code': 'UNK',
        'display': 'unknown'
    },
    {
        'system': 'http://terminology.hl7.org/CodeSystem/v3-NullFlavor',
        'code': 'ASKU',
        'display': 'asked but unknown'
    }
]

ethicity_choices = [
    {
        'system': 'urn:oid:2.16.840.1.113883.6.238',
        'code': '2135-2',
        'display': 'Hispanic or Latino'
    },
    {
        'system': 'urn:oid:2.16.840.1.113883.6.238',
        'code': '2186-5',
        'display': 'Non Hispanic or Latino'
    }
]


def generateUSCorePatient(config: dict) -> dict:
    '''Generate Patient Resource'''
    patient_resource = generatePatient(config)
    patient_resource['meta'] = {'profile': ['http://hl7.org/fhir/us/core/StructureDefinition/us-core-patient']}
    patient_resource['extension'] = [generateUSCoreRaceExtension(), generateUSCoreEthnicityExtension()]
    return patient_resource


def generateUSCoreRaceExtension() -> dict:
    '''Generate US Core Race Extension'''
    race_choice = random.choice(race_choices)

    us_core_race_ext_data = {
        'url': 'http://hl7.org/fhir/us/core/StructureDefinition/us-core-race',
        'extension': [
            {
                'url': 'ombCategory',
                'valueCoding': race_choice
            },
            {
                'url': 'text',
                'valueString': race_choice['display']
            }
        ]
    }

    us_core_race_extension = Extension(**us_core_race_ext_data)

    return dict(us_core_race_extension.dict())


def generateUSCoreEthnicityExtension() -> dict:
    '''Generate US Core Ethnicity Extension'''
    ethnicity_choice = random.choice(ethicity_choices)

    us_core_ethnicity_ext_data = {
        'url': 'http://hl7.org/fhir/us/core/StructureDefinition/us-core-ethnicity',
        'extension': [
            {
                'url': 'ombCategory',
                'valueCoding': ethnicity_choice
            },
            {
                'url': 'text',
                'valueString': ethnicity_choice['display']
            }
        ]
    }

    us_core_ethnicity_extension = Extension(**us_core_ethnicity_ext_data)

    return dict(us_core_ethnicity_extension.dict())
